@@ -1,6 +1,6 @@
 #
 # This file is part of VIRL 2
-# Copyright (c) 2019-2022, Cisco Systems, Inc.
+# Copyright (c) 2019-2023, Cisco Systems, Inc.
 # All rights reserved.
 #
 # Python bindings for the Cisco VIRL 2 Network Simulation Platform
@@ -18,7 +18,13 @@
 # limitations under the License.
 #
 
+from __future__ import annotations
+
 import io
+from typing import TYPE_CHECKING, Any, Optional
+
+if TYPE_CHECKING:
+    from .lab import Lab
 
 
 class PyatsNotInstalled(Exception):
@@ -30,16 +36,14 @@ class PyatsDeviceNotFound(Exception):
 
 
 class ClPyats:
-    def __init__(self, lab, hostname=None):
+    def __init__(self, lab: Lab, hostname: Optional[str] = None) -> None:
         """
         Creates a pyATS object that can be used to run commands
         against a device either in exec mode ``show version`` or in
         configuration mode ``interface gi0/0 \\n no shut``.
 
         :param lab: The lab which should be used with pyATS
-        :type lab: models.Lab
         :param hostname: Force hostname/ip and port for console terminal server
-        :type hostname: str
         :raises PyatsNotInstalled: when pyATS can not be found
         :raises PyatsDeviceNotFound: when the device can not be found
         """
@@ -52,25 +56,22 @@ class ClPyats:
             return
         else:
             self._pyats_installed = True
+        self._testbed: Any = None
+        self._connections: list[Any] = []
 
-        self._testbed = None
-        self._connections = []
-
-    def _check_pyats_installed(self):
+    def _check_pyats_installed(self) -> None:
         if not self._pyats_installed:
             raise PyatsNotInstalled
 
-    def sync_testbed(self, username, password):
+    def sync_testbed(self, username: str, password: str) -> None:
         """
         Syncs the testbed from the server. Note that this
         will always fetch the latest topology data from the server.
 
         :param username: the username that will be inserted into
             the testbed data
-        :type username: str
         :param password: the password that will be inserted into
             the testbed data
-        :type password: str
         """
         self._check_pyats_installed()
         from pyats.topology import loader
@@ -81,16 +82,13 @@ class ClPyats:
         data.devices.terminal_server.credentials.default.password = password
         self._testbed = data
 
-    def run_command(self, node_label, command):
+    def run_command(self, node_label: str, command: str) -> str:
         """
         Run a command on the device in `exec` mode.
 
         :param node_label: the label / title of the device
-        :type node_label: str
         :param command: the command to be run in exec mode
-        :type command: str
         :returns: The output from the device
-        :rtype: str
         """
         self._check_pyats_installed()
 
@@ -105,17 +103,14 @@ class ClPyats:
         self._connections.append(pyats_device)
         return pyats_device.execute(command, log_stdout=False)
 
-    def run_config_command(self, node_label, command):
+    def run_config_command(self, node_label: str, command: str) -> str:
         """
         Run a command on the device in `configure` mode. pyATS
         handles the change into `configure` mode automatically.
 
         :param node_label: the label / title of the device
-        :type node_label: str
         :param command: the command to be run in exec mode
-        :type command: str
         :returns: The output from the device
-        :rtype: str
         """
         self._check_pyats_installed()
 
@@ -130,7 +125,7 @@ class ClPyats:
         self._connections.append(pyats_device)
         return pyats_device.configure(command, log_stdout=False)
 
-    def cleanup(self):
+    def cleanup(self) -> None:
         """Cleans up the pyATS connections."""
         for pyats_device in self._connections:
             pyats_device.destroy()
