@@ -1,6 +1,6 @@
 #
 # This file is part of VIRL 2
-# Copyright (c) 2019-2022, Cisco Systems, Inc.
+# Copyright (c) 2019-2023, Cisco Systems, Inc.
 # All rights reserved.
 #
 # Python bindings for the Cisco VIRL 2 Network Simulation Platform
@@ -18,66 +18,65 @@
 # limitations under the License.
 #
 
+from __future__ import annotations
+
+from typing import TYPE_CHECKING, Optional
+
+if TYPE_CHECKING:
+    import httpx
+
 
 class GroupManagement:
-    def __init__(self, context):
-        self.ctx = context
+    def __init__(self, session: httpx.Client) -> None:
+        self._session = session
 
     @property
-    def base_url(self):
-        return self.ctx.base_url + "groups"
+    def base_url(self) -> str:
+        return "groups"
 
-    def groups(self):
+    def groups(self) -> list:
         """
         Get the list of available groups.
 
         :return: list of group objects
-        :rtype: list
         """
-        response = self.ctx.session.get(self.base_url)
-        response.raise_for_status()
-        return response.json()
+        return self._session.get(self.base_url).json()
 
-    def get_group(self, group_id):
+    def get_group(self, group_id: str) -> dict:
         """
         Gets info for the specified group.
 
         :param group_id: group uuid4
-        :type group_id: str
         :return: group object
-        :rtype: dict
         """
         url = self.base_url + "/{}".format(group_id)
-        response = self.ctx.session.get(url)
-        response.raise_for_status()
-        return response.json()
+        return self._session.get(url).json()
 
-    def delete_group(self, group_id):
+    def delete_group(self, group_id: str) -> None:
         """
         Deletes a group.
 
         :param group_id: group uuid4
-        :type group_id: str
         :return: None
         """
         url = self.base_url + "/{}".format(group_id)
-        response = self.ctx.session.delete(url)
-        response.raise_for_status()
+        self._session.delete(url)
 
-    def create_group(self, name, description="", members=None, labs=None):
+    def create_group(
+        self,
+        name: str,
+        description: str = "",
+        members: Optional[list[str]] = None,
+        labs: Optional[list[dict[str, str]]] = None,
+    ) -> dict:
         """
         Creates a group.
 
         :param name: group name
-        :type name: str
         :param description: group description
-        :type description: str
         :param members: group members
-        :type members: List[str]
         :param labs: group labs
-        :type labs: List[Dict[str, str]]
         :return: created group object
-        :rtype: dict
         """
         data = {
             "name": name,
@@ -85,30 +84,27 @@ class GroupManagement:
             "members": members or [],
             "labs": labs or [],
         }
-        response = self.ctx.session.post(self.base_url, json=data)
-        response.raise_for_status()
-        return response.json()
+        return self._session.post(self.base_url, json=data).json()
 
     def update_group(
-        self, group_id, name=None, description=None, members=None, labs=None
-    ):
+        self,
+        group_id: str,
+        name: Optional[str] = None,
+        description: Optional[str] = None,
+        members: Optional[list[str]] = None,
+        labs: Optional[list[dict[str, str]]] = None,
+    ) -> dict:
         """
         Updates a group.
 
         :param group_id: group uuid4
-        :type group_id: str
         :param name: new group name
-        :type name: str
         :param description: group description
-        :type description: str
         :param members: group members
-        :type members: List[str]
         :param labs: group labs
-        :type labs: List[Dict[str, str]]
         :return: updated group object
-        :rtype: dict
         """
-        data = {}
+        data: dict[str, str | list] = {}
         if name is not None:
             data["name"] = name
         if description is not None:
@@ -118,48 +114,34 @@ class GroupManagement:
         if labs is not None:
             data["labs"] = labs
         url = self.base_url + "/{}".format(group_id)
-        response = self.ctx.session.patch(url, json=data)
-        response.raise_for_status()
-        return response.json()
+        return self._session.patch(url, json=data).json()
 
-    def group_members(self, group_id):
+    def group_members(self, group_id: str) -> list[str]:
         """
         Gets group members.
 
         :param group_id: group uuid4
-        :type group_id: str
         :return: list of users associated with this group
-        :rtype: List[str]
         """
         url = self.base_url + "/{}/members".format(group_id)
-        response = self.ctx.session.get(url)
-        response.raise_for_status()
-        return response.json()
+        return self._session.get(url).json()
 
-    def group_labs(self, group_id):
+    def group_labs(self, group_id: str) -> list[str]:
         """
         Get the list of labs that are associated with this group.
 
         :param group_id: group uuid4
-        :type group_id: str
         :return: list of labs associated with this group
-        :rtype: List[str]
         """
         url = self.base_url + "/{}/labs".format(group_id)
-        response = self.ctx.session.get(url)
-        response.raise_for_status()
-        return response.json()
+        return self._session.get(url).json()
 
-    def group_id(self, group_name):
+    def group_id(self, group_name: str) -> str:
         """
         Get unique user uuid4.
 
         :param group_name: group name
-        :type group_name: str
         :return: group unique identifier
-        :rtype: str
         """
         url = self.base_url + "/{}/id".format(group_name)
-        response = self.ctx.session.get(url)
-        response.raise_for_status()
-        return response.json()
+        return self._session.get(url).json()
