@@ -1,6 +1,6 @@
 #
 # This file is part of VIRL 2
-# Copyright (c) 2019-2022, Cisco Systems, Inc.
+# Copyright (c) 2019-2023, Cisco Systems, Inc.
 # All rights reserved.
 #
 # Python bindings for the Cisco VIRL 2 Network Simulation Platform
@@ -18,31 +18,32 @@
 # limitations under the License.
 #
 
+from __future__ import annotations
+
 import json
 import logging
 import os
 import re
 import time
-from typing import NamedTuple, Optional, Dict, Union
-import urllib
 from functools import lru_cache
 from pathlib import Path
+from typing import Any, NamedTuple, Optional
 from urllib.parse import urljoin, urlsplit, urlunsplit
 
-import requests
-import urllib3
+import httpx
 
-from .exceptions import LabNotFound, InitializationError
+from .exceptions import InitializationError, LabNotFound
 from .models import (
-    Context,
-    Lab,
-    NodeImageDefinitions,
-    TokenAuth,
-    Licensing,
-    UserManagement,
-    SystemManagement,
     GroupManagement,
+    Lab,
+    Licensing,
+    NodeImageDefinitions,
+    ResourcePoolManagement,
+    SystemManagement,
+    TokenAuth,
+    UserManagement,
 )
+from .models.authentication import make_session
 
 _LOGGER = logging.getLogger(__name__)
 cached = lru_cache(maxsize=None)  # cache results forever
@@ -51,7 +52,7 @@ cached = lru_cache(maxsize=None)  # cache results forever
 class Version:
     __slots__ = ("version_str", "major", "minor", "patch")
 
-    def __init__(self, version_str: str):
+    def __init__(self, version_str: str) -> None:
         self.version_str = version_str
         version_tuple = self.parse_version_str(version_str)
         self.major = int(version_tuple[0])
@@ -59,7 +60,7 @@ class Version:
         self.patch = int(version_tuple[2])
 
     @staticmethod
-    def parse_version_str(version_str: str):
+    def parse_version_str(version_str: str) -> str:
         regex = r"^(\d+)\.(\d+)\.(\d+)(.*)$"
         res = re.findall(regex, version_str)
         if not res:
@@ -107,19 +108,19 @@ class Version:
     def __le__(self, other):
         return self == other or self < other
 
-    def major_differs(self, other):
+    def major_differs(self, other: Version) -> bool:
         return self.major != other.major
 
-    def minor_differs(self, other):
+    def minor_differs(self, other: Version) -> bool:
         return self.minor != other.minor
 
-    def minor_lt(self, other):
+    def minor_lt(self, other: Version) -> bool:
         return self.minor < other.minor
 
-    def patch_differs(self, other):
+    def patch_differs(self, other: Version) -> bool:
         return self.patch != other.patch
 
-    def minor_or_patch_differs(self, other):
+    def minor_or_patch_differs(self, other: Version) -> bool:
         return self.minor_differs(other) or self.patch_differs(other)
 
 
@@ -130,14 +131,14 @@ class ClientConfig(NamedTuple):
     url: Optional[str] = None
     username: Optional[str] = None
     password: Optional[str] = None
-    ssl_verify: bool = True
+    ssl_verify: bool | str = True
     allow_http: bool = False
     auto_sync: float = 1.0
     raise_for_auth_failure: bool = True
     convergence_wait_max_iter: int = 500
-    convergence_wait_time: Union[int, float] = 5
+    convergence_wait_time: int | float = 5
 
-    def make_client(self):
+    def make_client(self) -> ClientLibrary:
         client = ClientLibrary(
             url=self.url,
             username=self.username,
@@ -157,7 +158,7 @@ class ClientLibrary:
     """Python bindings for the REST API of a CML controller."""
 
     # current client version
-    VERSION = Version("2.4.1")
+    VERSION = Version("2.5.0")
     # list of Version objects
     INCOMPATIBLE_CONTROLLER_VERSIONS = [
         Version("2.0.0"),
@@ -172,15 +173,15 @@ class ClientLibrary:
 
     def __init__(
         self,
-        url=None,
-        username=None,
-        password=None,
-        ssl_verify=True,
-        raise_for_auth_failure=False,
-        allow_http=False,
-        convergence_wait_max_iter=500,
-        convergence_wait_time=5,
-    ):
+        url: Optional[str] = None,
+        username: Optional[str] = None,
+        password: Optional[str] = None,
+        ssl_verify: bool | str = True,
+        raise_for_auth_failure: bool = False,
+        allow_http: bool = False,
+        convergence_wait_max_iter: int = 500,
+        convergence_wait_time: int | float = 5,
+    ) -> None:
         """
         Initializes a ClientLibrary instance. Note that ssl_verify can
         also be a string that points to a cert (see class documentation).
@@ -188,73 +189,55 @@ class ClientLibrary:
         :param url: URL of controller. It's also possible to pass the
             URL via the ``VIRL2_URL`` environment variable. If no protocol
             scheme is provided, "https:" is used.
-        :type url: str
         :param username: Username of the user to authenticate
-        :type username: str
         :param password: Password of the user to authenticate
-        :type password: str
-        :param ssl_verify: SSL controller certificate verification
-        :type ssl_verify: bool
+        :param ssl_verify: Path SSL controller certificate, or True to load
+            from ``CA_BUNDLE`` environment variable, or False to disable
         :param raise_for_auth_failure: Raise an exception if unable
             to connect to controller (use for scripting scenarios)
-        :type raise_for_auth_failure: bool
         :param allow_http: If set, a https URL will not be enforced
-        :type allow_http: bool
         :param convergence_wait_max_iter: maximum number of iterations to wait for
             lab to converge
-        :type convergence_wait_max_iter: int
         :param convergence_wait_time: wait interval in seconds to wait during one
             iteration during lab convergence wait
-        :type convergence_wait_time: int
-        :raises InitializationError: If no URL is provided or
+        :raises InitializationError: If no URL is provided,
             authentication fails or host can't be reached
         """
 
-        url = self._environ_get(url, "VIRL2_URL")
+        url = self._environ_get("VIRL2_URL", url)
         if url is None or len(url) == 0:
             message = "no URL provided"
             raise InitializationError(message)
 
-        url, base_url = self._prepare_url(url, allow_http)
+        url, base_url = prepare_url(url, allow_http)
 
         # check environment for username
-        username = self._environ_get(username, "VIRL2_USER")
+        username = self._environ_get("VIRL2_USER", username)
         if username is None or len(username) == 0:
             message = "no username provided"
             raise InitializationError(message)
-        self.username = username
+        self.username: str = username
 
         # check environment for password
-        password = self._environ_get(password, "VIRL2_PASS")
+        password = self._environ_get("VIRL2_PASS", password)
         if password is None or len(password) == 0:
             message = "no password provided"
             raise InitializationError(message)
-        self.password = password
-
-        self._context = Context(base_url)
-        """
-        Within the Client Library context:
-        `requests.Session()` instance that can be used to send requests
-            to the controller.
-        `uuid.uuid4()` instance to uniquely identify this client library session.
-        `base_url` stores the base URL.
-        """
-
-        # http://docs.python-requests.org/en/master/user/advanced/#ssl-cert-verification
+        self.password: str = password
 
         if ssl_verify is True:
-            ssl_verify = self._environ_get(None, "CA_BUNDLE", True)
-        self.session.verify = ssl_verify
+            ssl_verify = self._environ_get("CA_BUNDLE", default=True)
 
         if ssl_verify is False:
             _LOGGER.warning("SSL Verification disabled")
-            urllib3.disable_warnings()
 
+        self._ssl_verify = ssl_verify
+        self._session = make_session(base_url, ssl_verify)
         # checks version from system_info against self.VERSION
         self.check_controller_version()
 
-        self.session.auth = TokenAuth(self)
-        # Note: session.auth is defined in the requests module to be of type AuthBase,
+        self._session.auth = TokenAuth(self)
+        # Note: session.auth is defined in the httpx module to be of type Auth,
         #  which has no logout function; this TokenAuth function will, therefore,
         #  not be visible to a type checker, causing warnings.
 
@@ -268,16 +251,25 @@ class ClientLibrary:
         self.convergence_wait_time = convergence_wait_time
 
         self.allow_http = allow_http
-        self.definitions = NodeImageDefinitions(self._context)
+        self.definitions = NodeImageDefinitions(self._session)
 
-        self.url = url
+        self.url: str = url
         self.raise_for_auth_failure = raise_for_auth_failure
-        self._labs: Dict[str, Lab] = {}
+        self._labs: dict[str, Lab] = {}
 
-        self.licensing = Licensing(context=self._context)
-        self.user_management = UserManagement(context=self._context)
-        self.group_management = GroupManagement(context=self._context)
-        self.system_management = SystemManagement(context=self._context)
+        self.licensing = Licensing(self._session)
+        self.user_management = UserManagement(self._session)
+        self.group_management = GroupManagement(self._session)
+        self.system_management = SystemManagement(
+            self._session,
+            auto_sync=self.auto_sync,
+            auto_sync_interval=self.auto_sync_interval,
+        )
+        self.resource_pool_management = ResourcePoolManagement(
+            self._session,
+            auto_sync=self.auto_sync,
+            auto_sync_interval=self.auto_sync_interval,
+        )
 
         try:
             self._make_test_auth_call()
@@ -294,65 +286,35 @@ class ClientLibrary:
             self.url,
             self.username,
             self.password,
-            self._context.session.verify,
+            self._ssl_verify,
             self.raise_for_auth_failure,
             self.allow_http,
         )
 
     def __str__(self):
-        return "{} URL: {}".format(self.__class__.__name__, self._context.base_url)
+        return "{} URL: {}".format(self.__class__.__name__, self._session.base_url)
 
-    def _prepare_url(self, url, allow_http):
-        # prepare the URL
-        try:
-            url_parts = urlsplit(url, "https")
-        except ValueError:
-            message = "invalid URL / hostname"
-            raise InitializationError(message)
-
-        # https://docs.python.org/3/library/urllib.parse.html
-        # Following the syntax specifications in RFC 1808, urlparse recognizes
-        # a netloc only if it is properly introduced by ‘//’. Otherwise, the
-        # input is presumed to be a relative URL and thus to start with
-        # a path component.
-        if len(url_parts.netloc) == 0:
-            try:
-                url_parts = urlsplit("//" + url, "https")
-            except ValueError:
-                message = "invalid URL / hostname"
-                raise InitializationError(message)
-
-        if not allow_http and url_parts.scheme == "http":
-            message = "invalid URL scheme (must be https)"
-            raise InitializationError(message)
-        if url_parts.scheme not in ("http", "https"):
-            message = "invalid URL scheme (should be https)"
-            raise InitializationError(message)
-        url = urlunsplit(url_parts)
-        base_url = urljoin(url, "api/v0/")
-        return url, base_url
-
-    def _make_test_auth_call(self):
+    def _make_test_auth_call(self) -> None:
         """Make a call to confirm auth works by using the "authok" endpoint."""
+        url = "authok"
         try:
-            url = self._base_url + "authok"
-            response = self.session.get(url)
-            response.raise_for_status()
-        except requests.HTTPError as exc:
-            response = exc.response
-            status_code = response.status_code
-            if status_code == requests.codes.forbidden:  # pylint: disable=E1101
+            self._session.get(url)
+        except httpx.HTTPStatusError as exc:
+            status_code = exc.response.status_code
+            if status_code == httpx.codes.FORBIDDEN:
                 message = (
                     "Unable to authenticate, please check your username and password"
                 )
                 raise InitializationError(message)
             raise
-        except requests.exceptions.ConnectionError as exc:
+        except httpx.HTTPError as exc:
             # TODO: subclass InitializationError
             raise InitializationError(exc)
 
     @staticmethod
-    def _environ_get(value, key, default=None):
+    def _environ_get(
+        key: str, value: Optional[Any] = None, default: Optional[Any] = None
+    ) -> Optional[Any]:
         """If the value is not set yet, fetch it from environment or return default."""
         if value is None:
             value = os.environ.get(key)
@@ -363,111 +325,89 @@ class ClientLibrary:
         return value
 
     @property
-    def session(self):
+    def session(self) -> httpx.Client:
         """
-        Returns the used Requests session object.
+        Returns the used `httpx` client session object.
 
-        :returns: The Requests session object
-        :rtype: requests.Session
+        :returns: The session object
         """
-        return self._context.session
+        return self._session
 
-    @property
-    def _base_url(self):
-        """
-        Returns the base URL to access the controller.
-
-        :returns: The base URL
-        :rtype: str
-        """
-        return self._context.base_url
-
-    def logout(self, clear_all_sessions=False):
+    def logout(self, clear_all_sessions: bool = False) -> None:
         """
         Invalidate current token.
         """
-        # noinspection PyUnresolvedReferences
-        return self.session.auth.logout(clear_all_sessions=clear_all_sessions)
-        # See session.auth assignment above for noinspection reasoning.
+        return self._session.auth.logout(  # type: ignore
+            clear_all_sessions=clear_all_sessions
+        )
 
-    def get_host(self):
+    def get_host(self) -> str:
         """
         Returns the hostname of the session to the controller.
 
         :returns: A hostname
-        :rtype: str
         """
-        split_url = urllib.parse.urlsplit(self._base_url)
-        return split_url.hostname
+        return self.session.base_url.host
 
-    def system_info(self):
+    def system_info(self) -> dict:
         """
         Get information about the system where the application runs.
         Can be called without authentication.
 
         :returns: system information json
-        :rtype: dict
         """
-        response = self.session.get(urljoin(self._base_url, "system_information"))
-        response.raise_for_status()
-        return response.json()
+        url = "system_information"
+        return self._session.get(url).json()
 
-    def check_controller_version(self, controller_version=None):
+    def check_controller_version(self) -> None:
         """
         Checks remote controller version against current client version
         (specified in self.VERSION) and against controller version
         blacklist (specified in self.INCOMPATIBLE_CONTROLLER_VERSIONS) and
         raises exception if versions are incompatible or prints warning
         if client minor version is lower than controller minor version.
-
-        :type controller_version: str
-        :rtype: None
         """
-        controller_version = (
-            controller_version or self.system_info().get("version", "").split("-")[0]
-        )
+        controller_version = self.system_info().get("version", "").split("-")[0]
 
         # are we running against a test version?
         if controller_version == "testing":
             _LOGGER.warning("testing version detected!")
             return
 
-        controller_version = Version(controller_version)
-        if controller_version in self.INCOMPATIBLE_CONTROLLER_VERSIONS:
+        controller_version_obj = Version(controller_version)
+        if controller_version_obj in self.INCOMPATIBLE_CONTROLLER_VERSIONS:
             raise InitializationError(
                 "Controller version {} is marked incompatible! "
                 "List of versions marked explicitly as incompatible: {}.".format(
-                    controller_version, self.INCOMPATIBLE_CONTROLLER_VERSIONS
+                    controller_version_obj, self.INCOMPATIBLE_CONTROLLER_VERSIONS
                 )
             )
-        if self.VERSION.major_differs(controller_version):
+        if self.VERSION.major_differs(controller_version_obj):
             raise InitializationError(
                 "Major version mismatch. Client {}, controller {}.".format(
-                    self.VERSION, controller_version
+                    self.VERSION, controller_version_obj
                 )
             )
-        if self.VERSION.minor_differs(controller_version) and self.VERSION.minor_lt(
-            controller_version
+        if self.VERSION.minor_differs(controller_version_obj) and self.VERSION.minor_lt(
+            controller_version_obj
         ):
             _LOGGER.warning(
                 "Please ensure the client version is compatible with the controller"
                 " version. Client %s, controller %s.",
                 self.VERSION,
-                controller_version,
+                controller_version_obj,
             )
 
-    def is_system_ready(self, wait=False, max_wait=60, sleep=5):
+    def is_system_ready(
+        self, wait: bool = False, max_wait: int = 60, sleep: int = 5
+    ) -> bool:
         """
         Reports whether the system is ready or not.
 
         :param wait: if this is true, the call will block until it's ready
-        :type wait: bool
         :param max_wait: maximum time to wait, in seconds
-        :type wait: int
         :param sleep: time to wait between tries, in seconds
-        :type wait: int
         :returns: ready state
-        :rtype: bool
         """
         loops = 1
         if wait:
@@ -476,57 +416,59 @@ class ClientLibrary:
         while not ready and loops > 0:
             try:
                 result = self.system_info()
-                ready = result.get("ready")
-            except requests.HTTPError:
-                # 502 Bad Gateway is expected and hints
-                # that system is not ready - no need to
-                # raise - just wait
-                ready = False
+                ready = bool(result.get("ready"))
+            except httpx.HTTPStatusError as exc:
+                if exc.response.status_code == httpx.codes.BAD_GATEWAY:
+                    # 502 Bad Gateway is expected and hints
+                    # that system is not ready - no need to
+                    # raise - just wait
+                    ready = False
+                else:
+                    raise
             if not ready and loops > 0:
                 time.sleep(sleep)
             loops -= 1
         return ready
 
     @staticmethod
-    def is_virl_1x(path: Path):
+    def is_virl_1x(path: Path) -> bool:
         if path.suffix == ".virl":
             return True
         return False
 
-    def import_lab(self, topology, title=None, offline=False, virl_1x=False):
+    def import_lab(
+        self,
+        topology: str,
+        title: Optional[str] = None,
+        offline: bool = False,
+        virl_1x: bool = False,
+    ) -> Lab:
         """
         Imports an existing topology from a string.
 
         :param topology: Topology representation as a string
-        :type topology: str
         :param title: Title of the lab (optional)
-        :type title: str (otherwise ignored)
         :param offline: whether the ClientLibrary should import the
             lab locally.  The topology parameter has to be a JSON string
             in this case.  This can not be XML or YAML representation of
             the lab.  Compatible JSON from `GET /labs/{lab_id}/topology`.
-        :type offline: bool
         :param virl_1x: Is this old virl-1x topology format (default=False)
-        :type virl_1x: bool
         :returns: A Lab instance
-        :rtype: models.Lab
         :raises ValueError: if there's no lab ID in the API response
-        :raises requests.exceptions.HTTPError: if there was a transport error
+        :raises httpx.HTTPError: if there was a transport error
         """
         # TODO: refactor to return the local lab, and sync it,
-        # if already exists in self._labs
+        #  if already exists in self._labs
         if offline:
             lab_id = "offline_lab"
         else:
             if virl_1x:
-                url = "{}import/virl-1x".format(self._base_url)
+                url = "import/virl-1x"
             else:
-                url = "{}import".format(self._base_url)
-            if isinstance(title, str):
+                url = "import"
+            if title is not None:
                 url = "{}?title={}".format(url, title)
-            response = self.session.post(url, data=topology)
-            response.raise_for_status()
-            result = response.json()
+            result = self._session.post(url, content=topology).json()
             lab_id = result.get("id")
             if lab_id is None:
                 raise ValueError("No lab ID returned!")
@@ -534,11 +476,12 @@ class ClientLibrary:
         lab = Lab(
             title,
             lab_id,
-            self._context,
+            self._session,
             self.username,
             self.password,
             auto_sync=self.auto_sync,
             auto_sync_interval=self.auto_sync_interval,
+            resource_pool_manager=self.resource_pool_management,
         )
 
         if offline:
@@ -552,16 +495,13 @@ class ClientLibrary:
         self._labs[lab_id] = lab
         return lab
 
-    def import_lab_from_path(self, path, title=None):
+    def import_lab_from_path(self, path: str, title: Optional[str] = None) -> Lab:
         """
         Imports an existing topology from a file / path.
 
         :param path: Topology filename / path
-        :type path: str
         :param title: Title of the lab
-        :type title: str
         :returns: A Lab instance
-        :rtype: models.Lab
         """
         topology_path = Path(path)
         if not topology_path.exists():
@@ -573,40 +513,32 @@ class ClientLibrary:
             topology, title=title, virl_1x=self.is_virl_1x(topology_path)
         )
 
-    def get_sample_labs(self):
+    def get_sample_labs(self) -> dict[str, dict]:
         """
         Returns a dictionary with information about all sample labs available on host.
 
-        :returns: A dictionary of sample lab information
-        :rtype: dict
+        :returns: A dictionary of sample lab information, where keys are titles
         """
-        url = urljoin(self._base_url, "sample/labs")
-        response = self.session.get(url)
-        response.raise_for_status()
-        return response.json()
+        url = "sample/labs"
+        return self._session.get(url).json()
 
-    def import_sample_lab(self, title):
+    def import_sample_lab(self, title: str) -> Lab:
         """
         Imports a built-in sample lab.
 
         :param title: sample lab name, as returned by get_sample_labs
-        :type title: str
         :returns: a Lab instance
-        :rtype: models.Lab
         """
 
-        url = urljoin(self._base_url, "sample/labs/" + title)
-        response = self.session.put(url)
-        response.raise_for_status()
-        lab_id = response.json()
+        url = "sample/labs/" + title
+        lab_id = self._session.put(url).json()
         return self.join_existing_lab(lab_id)
 
-    def all_labs(self, show_all=False):
+    def all_labs(self, show_all: bool = False) -> list[Lab]:
         """
         Retrieves a list of all defined labs.
 
         :param show_all: Whether to get only labs owned by the admin or all user labs
-        :type show_all: bool
         :returns: A list of lab objects
         :rtype: list[models.Lab]
         """
@@ -614,10 +546,7 @@ class ClientLibrary:
         url = "labs"
         if show_all:
             url += "?show_all=true"
-        url = urljoin(self._base_url, url)
-        response = self.session.get(url)
-        response.raise_for_status()
-        lab_ids = response.json()
+        lab_ids = self._session.get(url).json()
         result = []
         for lab_id in lab_ids:
             lab = self.join_existing_lab(lab_id)
@@ -625,17 +554,17 @@ class ClientLibrary:
 
         return result
 
-    def local_labs(self):
+    def local_labs(self) -> list[Lab]:
         # TODO: first sync with controller to pull all possible labs
         return [lab for lab in self._labs.values()]
 
-    def get_local_lab(self, lab_id):
+    def get_local_lab(self, lab_id: str) -> Lab:
         try:
             return self._labs[lab_id]
         except KeyError:
             raise LabNotFound(lab_id)
 
-    def create_lab(self, title=None):
+    def create_lab(self, title: Optional[str] = None) -> Lab:
         """
         Creates an empty lab with the given name. If no name is given, then
         the created lab ID is set as the name.
@@ -651,32 +580,29 @@ class ClientLibrary:
             lab.create_node("r1", "iosv", 50, 100)
 
         :param title: The Lab name (or title)
-        :type title: str
         :returns: A Lab instance
-        :rtype: models.Lab
         """
-        url = self._base_url + "labs"
+        url = "labs"
         params = {"title": title}
-        response = self.session.post(url, params=params)
-        response.raise_for_status()
-        result = response.json()
+        result = self._session.post(url, params=params).json()
         # TODO: server generated title not loaded
         lab_id = result["id"]
         lab = Lab(
             title or lab_id,
             lab_id,
-            self._context,
+            self._session,
             self.username,
             self.password,
             auto_sync=self.auto_sync,
             auto_sync_interval=self.auto_sync_interval,
             wait_max_iterations=self.convergence_wait_max_iter,
             wait_time=self.convergence_wait_time,
+            resource_pool_manager=self.resource_pool_management,
         )
         self._labs[lab_id] = lab
         return lab
 
-    def remove_lab(self, lab_id):
+    def remove_lab(self, lab_id: str) -> None:
         """
         Use carefully, it removes a given lab::
 
@@ -685,12 +611,11 @@ class ClientLibrary:
         :param lab_id: The lab ID to be removed.
         :type lab_id: str
         """
-        url = self._base_url + "labs/{}".format(lab_id)
-        response = self.session.delete(url)
-        response.raise_for_status()
+        url = "labs/{}".format(lab_id)
+        self._session.delete(url)
         self._labs.pop(lab_id, None)
 
-    def join_existing_lab(self, lab_id, sync_lab=True):
+    def join_existing_lab(self, lab_id: str, sync_lab: bool = True) -> Lab:
         """
         Creates a new ClientLibrary lab instance that is retrieved
         from the controller.
@@ -704,11 +629,8 @@ class ClientLibrary:
             lab = client_library.join_existing_lab("2e6a18")
 
         :param lab_id: The lab ID to be joined.
-        :type lab_id: str
         :param sync_lab: Synchronize changes.
-        :type sync_lab: bool
         :returns: A Lab instance
-        :rtype: models.Lab
         """
         # TODO: check if lab exists through REST call
         title = None
@@ -716,11 +638,12 @@ class ClientLibrary:
         lab = Lab(
             title,
             lab_id,
-            self._context,
+            self._session,
             self.username,
             self.password,
             auto_sync=self.auto_sync,
             auto_sync_interval=self.auto_sync_interval,
+            resource_pool_manager=self.resource_pool_management,
         )
         if sync_lab:
             lab.sync()
@@ -728,56 +651,43 @@ class ClientLibrary:
         self._labs[lab_id] = lab
         return lab
 
-    def get_diagnostics(self):
+    def get_diagnostics(self) -> dict:
         """
         Returns the controller diagnostic data as JSON
 
         :returns: diagnostic data
         :rtype: dict
         """
-        url = self._base_url + "diagnostics"
-        response = self.session.get(url)
-        response.raise_for_status()
-        return response.json()
+        url = "diagnostics"
+        return self._session.get(url).json()
 
-    def get_system_health(self):
+    def get_system_health(self) -> dict:
         """
         Returns the controller system health data as JSON
 
         :returns: system health data
-        :rtype: dict
         """
-        url = self._base_url + "system_health"
-        response = self.session.get(url)
-        response.raise_for_status()
-        return response.json()
+        url = "system_health"
+        return self._session.get(url).json()
 
-    def get_system_stats(self):
+    def get_system_stats(self) -> dict:
         """
         Returns the controller resource statistics as JSON
 
         :returns: system resource statistics
-        :rtype: dict
         """
-        url = self._base_url + "system_stats"
-        response = self.session.get(url)
-        response.raise_for_status()
-        return response.json()
+        url = "system_stats"
+        return self._session.get(url).json()
 
-    def find_labs_by_title(self, title):
+    def find_labs_by_title(self, title: str) -> list[Lab]:
         """
         Return a list of labs which match the given title.
 
         :param title: The title to search for
-        :type title: str
         :returns: A list of lab objects which match the title specified
-        :rtype: list[models.Lab]
         """
-        url = self._base_url + "populate_lab_tiles"
-        response = self.session.get(url)
-        response.raise_for_status()
-
-        resp = response.json()
+        url = "populate_lab_tiles"
+        resp = self._session.get(url).json()
 
         # populate_lab_tiles response has been changed in 2.1
         # if it doesn't have a key "lab_tiles" then it's <2.1
@@ -798,7 +708,7 @@ class ClientLibrary:
 
         return matched_labs
 
-    def get_lab_list(self, show_all=False):
+    def get_lab_list(self, show_all: bool = False) -> list[str]:
         """
         Returns list of all lab IDs.
 
@@ -810,7 +720,35 @@ class ClientLibrary:
         url = "labs"
         if show_all:
             url += "?show_all=true"
-        url = urljoin(self._base_url, url)
-        response = self.session.get(url)
-        response.raise_for_status()
-        return response.json()
+        return self._session.get(url).json()
+
+
+def prepare_url(url: str, allow_http: bool) -> tuple[str, str]:
+    # prepare the URL
+    try:
+        url_parts = urlsplit(url, "https")
+    except ValueError:
+        message = "invalid URL / hostname"
+        raise InitializationError(message)
+
+    # https://docs.python.org/3/library/urllib.parse.html
+    # Following the syntax specifications in RFC 1808, urlparse recognizes
+    # a netloc only if it is properly introduced by ‘//’. Otherwise, the
+    # input is presumed to be a relative URL and thus to start with
+    # a path component.
+    if len(url_parts.netloc) == 0:
+        try:
+            url_parts = urlsplit("//" + url, "https")
+        except ValueError:
+            message = "invalid URL / hostname"
+            raise InitializationError(message)
+
+    if not allow_http and url_parts.scheme == "http":
+        message = "invalid URL scheme (must be https)"
+        raise InitializationError(message)
+    if url_parts.scheme not in ("http", "https"):
+        message = "invalid URL scheme (should be https)"
+        raise InitializationError(message)
+    url = urlunsplit(url_parts)
+    base_url = urljoin(url, "api/v0/")
+    return url, base_url
